/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.factories;

import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.factories.annotations.DefaultFactoryFor;
import org.jboss.cache.interceptors.*;

import java.util.ArrayList;
import java.util.List;

/**
 * Factory class that builds an interceptor chain based on cache configuration.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@DefaultFactoryFor(classes = Interceptor.class)
public class InterceptorChainFactory extends ComponentFactory
{
   /**
    * Note - this method used to return a singleton instance, and since 2.1.0 returns a new instance.  The method is
    * deprecated and you should use the no-arg constructor to create a new instance of this factory.
    *
    * @return a NEW instance of this class.
    */
   @Deprecated
   public static InterceptorChainFactory getInstance()
   {
      return new InterceptorChainFactory();
   }

   public static Interceptor setLastInterceptorPointer(Interceptor first, Interceptor last)
   {
      Interceptor i = first;
      while (i != null)
      {
         i.setLast(last);
         i = i.getNext();
      }
      return first;
   }

   private Interceptor createInterceptor(Class<? extends Interceptor> clazz) throws IllegalAccessException, InstantiationException
   {
      Interceptor i = componentRegistry.getComponent(clazz.getName(), clazz);
      if (i == null)
      {
         i = clazz.newInstance();
//         componentRegistry.wireDependencies(i);
         // add this interceptor as a NAMED component into the registry
         componentRegistry.registerComponent(clazz.getName(), i, clazz);
      }
      else
      {
         // wipe next/last chaining!!
         i.setLast(null);
         i.setNext(null);
      }
      return i;
   }

   /**
    * Crates and adds an interceptor at the end of the chain
    */
   private void addInterceptor(Interceptor first, Class<? extends Interceptor> clazz) throws IllegalAccessException, InstantiationException
   {
      if (first == null) return;
      while (first.getNext() != null) first = first.getNext();
      first.setNext(createInterceptor(clazz));
   }

   /**
    * Adds an interceptor at the end of the chain
    */
   private void addInterceptor(Interceptor first, Interceptor i)
   {
      if (first == null) return;
      while (first.getNext() != null) first = first.getNext();
      first.setNext(i);
   }

   public Interceptor buildInterceptorChain() throws IllegalAccessException, InstantiationException, ClassNotFoundException
   {
      boolean optimistic = configuration.isNodeLockingOptimistic();
      // load the icInterceptor first
      Interceptor first = setFirstInterceptor(InvocationContextInterceptor.class);

      // load the cache management interceptor next
      if (configuration.getExposeManagementStatistics()) addInterceptor(first, CacheMgmtInterceptor.class);

      // load the tx interceptor
      addInterceptor(first, TxInterceptor.class);
      if (configuration.isUseLazyDeserialization()) addInterceptor(first, MarshalledValueInterceptor.class);
      addInterceptor(first, NotificationInterceptor.class);

      switch (configuration.getCacheMode())
      {
         case REPL_SYNC:
         case REPL_ASYNC:
            addInterceptor(first, optimistic ? OptimisticReplicationInterceptor.class : ReplicationInterceptor.class);
            break;
         case INVALIDATION_SYNC:
         case INVALIDATION_ASYNC:
            addInterceptor(first, InvalidationInterceptor.class);
            break;
         case LOCAL:
            //Nothing...
      }

      if (!optimistic)
      {
         addInterceptor(first, PessimisticLockInterceptor.class);
//         addInterceptor(first, UnlockInterceptor.class);
      }

      if (isUsingCacheLoaders())
      {
         if (configuration.getCacheLoaderConfig().isPassivation())
         {
            addInterceptor(first, ActivationInterceptor.class);
            addInterceptor(first, PassivationInterceptor.class);

         }
         else
         {
            addInterceptor(first, CacheLoaderInterceptor.class);
            addInterceptor(first, CacheStoreInterceptor.class);

         }
      }

      if (isUsingBuddyReplication()) addInterceptor(first, DataGravitatorInterceptor.class);

      if (optimistic)
      {
         addInterceptor(first, OptimisticLockingInterceptor.class);
         addInterceptor(first, OptimisticValidatorInterceptor.class);
         addInterceptor(first, OptimisticCreateIfNotExistsInterceptor.class);
      }
      // eviction interceptor to come before the optimistic node interceptor
      if (configuration.getEvictionConfig() != null && configuration.getEvictionConfig().isValidConfig())
         addInterceptor(first, EvictionInterceptor.class);

      if (optimistic) addInterceptor(first, OptimisticNodeInterceptor.class);

      Interceptor callInterceptor = createInterceptor(CallInterceptor.class);
      addInterceptor(first, callInterceptor);
      return setLastInterceptorPointer(first, callInterceptor);
   }

   private boolean isUsingCacheLoaders()
   {
      return configuration.getCacheLoaderConfig() != null && !configuration.getCacheLoaderConfig().getIndividualCacheLoaderConfigs().isEmpty();
   }

   private boolean isUsingBuddyReplication()
   {
      return configuration.getBuddyReplicationConfig() != null && configuration.getBuddyReplicationConfig().isEnabled() && configuration.getCacheMode() != Configuration.CacheMode.LOCAL;
   }

   public Interceptor setFirstInterceptor(Class<? extends Interceptor> clazz) throws IllegalAccessException, InstantiationException
   {
      return setFirstInterceptor(createInterceptor(clazz));
   }

   public Interceptor setFirstInterceptor(Interceptor i)
   {
      componentRegistry.registerComponent(Interceptor.class.getName(), i, Interceptor.class);
      return i;
   }


   public static List<Interceptor> asList(Interceptor interceptor)
   {
      if (interceptor == null)
      {
         return null;
      }
      int num = 1;
      Interceptor tmp = interceptor;
      while ((tmp = tmp.getNext()) != null)
      {
         num++;
      }
      List<Interceptor> retval = new ArrayList<Interceptor>(num);
      tmp = interceptor;
      num = 0;
      do
      {
         retval.add(tmp);
         tmp = tmp.getNext();
      }
      while (tmp != null);
      return retval;
   }

   /**
    * The same as {@link #correctInterceptorChaining(java.util.List)}
    * except that it takes in the first interceptor in the chain instead of a list of interceptors, and traverses the chain to
    * generate a list.  Identical to calling
    * <pre>
    *    InterceptorChainFactory.correctInterceptorChaining( InterceptorChainFactory.asList(first), configuration, componentRegistry );
    * </pre>
    *
    * @param first first interceptor in the chain to correct.
    * @return the first interceptor in the chain.
    */
   public Interceptor correctInterceptorChaining(Interceptor first)
   {
      return correctInterceptorChaining(asList(first));
   }

   /**
    * "Fixes" the next() and last() pointers for each interceptor, based on the order presented in the list passed in, and
    * also re-assesses dependencies for each interceptor, injecting dependencies accordingingly.
    *
    * @param interceptors interceptor chain to correct
    * @return the first interceptor in the chain.
    */
   public Interceptor correctInterceptorChaining(List<Interceptor> interceptors)
   {

      Interceptor first = null, last = null;

      for (Interceptor next : interceptors)
      {
         if (first == null)
         {
            first = last = next;
            continue;
         }
         last.setNext(next);
         last = next;
      }

      if (last != null) last.setNext(null);

      // now set the 'last' pointer.
      Interceptor i = setLastInterceptorPointer(first, last);
      // re-register this interceptor
      componentRegistry.registerComponent(Interceptor.class.getName(), i, Interceptor.class);
      return i;
   }

   /**
    * Initialises the interceptors with a CacheSPI instance.  Essentially calls setCache().  This is done AFTER the interceptor chain
    * is created, and when Cache.start() is called, so the interceptors get a fully initialised cache.
    *
    * @param componentName
    */
//   public void initialiseInterceptors(Interceptor firstInterceptor, CacheSPI cache)
//   {
//      for (Interceptor i : asList(firstInterceptor)) i.setCache(cache);
//   }
   @Override
   @SuppressWarnings("unchecked")
   protected <T> T construct(String componentName, Class<T> componentType)
   {
      try
      {
         return (T) buildInterceptorChain();
      }
      catch (Exception e)
      {
         throw new ConfigurationException("Unable to build interceptor chain", e);
      }
   }

   public static InterceptorChainFactory getInstance(ComponentRegistry componentRegistry, Configuration configuration)
   {
      InterceptorChainFactory icf = new InterceptorChainFactory();
      icf.componentRegistry = componentRegistry;
      icf.configuration = configuration;
      return icf;
   }
}
