/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.HashMap;
import java.util.Map;

/**
 * Captures cache management statistics
 *
 * @author Jerry Gauthier
 * @version $Id: CacheMgmtInterceptor.java 4968 2008-01-03 19:17:09Z manik.surtani@jboss.com $
 */
public class CacheMgmtInterceptor extends MethodDispacherInterceptor implements CacheMgmtInterceptorMBean
{
   private long m_hit_times = 0;
   private long m_miss_times = 0;
   private long m_store_times = 0;
   private long m_hits = 0;
   private long m_misses = 0;
   private long m_stores = 0;
   private long m_evictions = 0;
   private long m_start = System.currentTimeMillis();
   private long m_reset = m_start;

   public CacheMgmtInterceptor()
   {
      initLogger();
   }

   @Override
   protected boolean skipMethodCall(InvocationContext ctx)
   {
      return !getStatisticsEnabled();
   }

   @Override
   protected Object handleEvictMethod(InvocationContext ctx, Fqn fqn) throws Throwable
   {
      Object returnValue = nextInterceptor(ctx);
      m_evictions++;
      return returnValue;
   }

   @Override
   protected Object handleEvictVersionedNodeMethod(InvocationContext ctx, Fqn fqn, DataVersion dataVersion) throws Throwable
   {
      Object returnValue = nextInterceptor(ctx);
      m_evictions++;
      return returnValue;
   }

   @Override
   protected Object handleGetKeyValueMethod(InvocationContext ctx, Fqn fqn, Object key, boolean sendNodeEvent) throws Throwable
   {
      long t1 = System.currentTimeMillis();
      Object retval = super.nextInterceptor(ctx);
      long t2 = System.currentTimeMillis();
      if (retval == null)
      {
         m_miss_times = m_miss_times + (t2 - t1);
         m_misses++;
      }
      else
      {
         m_hit_times = m_hit_times + (t2 - t1);
         m_hits++;
      }
      return retval;
   }

   @Override
   protected Object handlePutDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Map data, boolean createUndoOps) throws Throwable
   {
      return handlePutData(ctx, data);
   }

   @Override
   protected Object handlePutDataEraseMethod(InvocationContext ctx, GlobalTransaction gt, Fqn fqn, Map newData, boolean createUndoOps, boolean eraseContents) throws Throwable
   {
      return handlePutData(ctx, newData);
   }

   private Object handlePutData(InvocationContext ctx, Map data)
         throws Throwable
   {
      long t1 = System.currentTimeMillis();
      Object retval = nextInterceptor(ctx);
      long t2 = System.currentTimeMillis();

      if (data != null && data.size() > 0)
      {
         m_store_times = m_store_times + (t2 - t1);
         m_stores = m_stores + data.size();
      }
      return retval;
   }

   @Override
   protected Object handlePutForExternalReadMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Object key, Object value) throws Throwable
   {
      return handlePutExternalAndKeyValue(ctx);
   }

   @Override
   protected Object handlePutKeyValueMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps) throws Throwable
   {
      return handlePutExternalAndKeyValue(ctx);
   }

   private Object handlePutExternalAndKeyValue(InvocationContext ctx)
         throws Throwable
   {
      long t1 = System.currentTimeMillis();
      Object retval = nextInterceptor(ctx);
      long t2 = System.currentTimeMillis();
      m_store_times = m_store_times + (t2 - t1);
      m_stores++;
      return retval;
   }

   public long getHits()
   {
      return m_hits;
   }

   public long getMisses()
   {
      return m_misses;
   }

   public long getStores()
   {
      return m_stores;
   }

   public long getEvictions()
   {
      return m_evictions;
   }

   public double getHitMissRatio()
   {
      double total = m_hits + m_misses;
      if (total == 0)
         return 0;
      return (m_hits / total);
   }

   public double getReadWriteRatio()
   {
      if (m_stores == 0)
         return 0;
      return (((double) (m_hits + m_misses) / (double) m_stores));
   }

   public long getAverageReadTime()
   {
      long total = m_hits + m_misses;
      if (total == 0)
         return 0;
      return (m_hit_times + m_miss_times) / total;
   }

   public long getAverageWriteTime()
   {
      if (m_stores == 0)
         return 0;
      return (m_store_times) / m_stores;
   }

   public int getNumberOfAttributes()
   {
      return cache.getNumberOfAttributes();
   }

   public int getNumberOfNodes()
   {
      return cache.getNumberOfNodes();
   }

   public long getElapsedTime()
   {
      return (System.currentTimeMillis() - m_start) / 1000;
   }

   public long getTimeSinceReset()
   {
      return (System.currentTimeMillis() - m_reset) / 1000;
   }

   @Override
   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("Hits", m_hits);
      retval.put("Misses", m_misses);
      retval.put("Stores", m_stores);
      retval.put("Evictions", m_evictions);
      retval.put("NumberOfAttributes", cache.getNumberOfAttributes());
      retval.put("NumberOfNodes", cache.getNumberOfNodes());
      retval.put("ElapsedTime", getElapsedTime());
      retval.put("TimeSinceReset", getTimeSinceReset());
      retval.put("AverageReadTime", getAverageReadTime());
      retval.put("AverageWriteTime", getAverageWriteTime());
      retval.put("HitMissRatio", getHitMissRatio());
      retval.put("ReadWriteRatio", getReadWriteRatio());
      return retval;
   }

   @Override
   public void resetStatistics()
   {
      m_hits = 0;
      m_misses = 0;
      m_stores = 0;
      m_evictions = 0;
      m_hit_times = 0;
      m_miss_times = 0;
      m_store_times = 0;
      m_reset = System.currentTimeMillis();
   }
}

