/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.optimistic.DefaultDataVersion;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.util.concurrent.ConcurrentHashSet;
import org.jgroups.Address;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Replication interceptor for the optimistically locked interceptor chain.  Responsible for replicating
 * state to remote nodes.  Unlike it's cousin, the {@link org.jboss.cache.interceptors.ReplicationInterceptor}, this interceptor
 * only deals with transactional calls.  Just like all things to do with Optimistic Locking, it is a requirement that
 * everything is done in a transaction and the transaction context is available via {@link org.jboss.cache.InvocationContext#getTransaction()}
 * and {@link org.jboss.cache.InvocationContext#getGlobalTransaction()}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:stevew@jofti.com">Steve Woodcock (stevew@jofti.com)</a>
 */
public class OptimisticReplicationInterceptor extends BaseRpcInterceptor
{

   // record of local broacasts - so we do not broadcast rollbacks/commits that resuted from
   // local prepare failures
   // we really just need a set here, but concurrent CopyOnWriteArraySet has poor performance when writing.
   private final Set<GlobalTransaction> broadcastTxs = new ConcurrentHashSet<GlobalTransaction>();

   public OptimisticReplicationInterceptor()
   {
      initLogger();
   }

   @Override
   protected boolean skipMethodCall(InvocationContext ctx)
   {
      // bypass for buddy group org metod calls.
      if (MethodDeclarations.isBuddyGroupOrganisationMethod(ctx.getMethodCall().getMethodId()))
         return true;
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null && optionOverride.isCacheModeLocal() && ctx.getTransaction() == null)
      {
         // skip replication!!
         log.debug("Skipping replication for this call as cache mode is local, forced via an option override.");
         return true;
      }
      return false;
   }

   @Override
   protected Object handleOptimisticPrepareMethod(InvocationContext ctx, GlobalTransaction gtx, List modifications, Map data, Address address, boolean onePhaseCommit) throws Throwable
   {
      // pass up the chain.
      Object retval = nextInterceptor(ctx);
      if (!skipReplicationOfTransactionMethod(ctx))
      {
         gtx = getGlobalTransaction(ctx);

         if (!gtx.isRemote() && ctx.isOriginLocal())
         {
            // replicate the prepare call.
            broadcastPrepare(ctx.getMethodCall(), gtx, ctx);
         }
      }
      return retval;
   }

   @Override
   protected Object handleCommitMethod(InvocationContext ctx, GlobalTransaction gtx) throws Throwable
   {
      //lets broadcast the commit first
      Throwable remoteCommitException = null;
      gtx = getGlobalTransaction(ctx);
      if (!gtx.isRemote() && ctx.isOriginLocal() && broadcastTxs.contains(gtx))
      {
         //we dont do anything
         try
         {
            if (!skipReplicationOfTransactionMethod(ctx)) broadcastCommit(gtx, ctx);
         }
         catch (Throwable t)
         {
            log.error("A problem occurred with remote commit", t);
            remoteCommitException = t;
         }
      }

      Object retval = nextInterceptor(ctx);
      if (remoteCommitException != null)
      {
         throw remoteCommitException;
      }
      return retval;
   }

   @Override
   protected Object handleRollbackMethod(InvocationContext ctx, GlobalTransaction gtx) throws Throwable
   {
      //    lets broadcast the rollback first
      gtx = getGlobalTransaction(ctx);
      Throwable remoteRollbackException = null;
      if (!gtx.isRemote() && ctx.isOriginLocal() && broadcastTxs.contains(gtx))
      {
         //we dont do anything
         try
         {
            if (!skipReplicationOfTransactionMethod(ctx)) broadcastRollback(gtx, ctx);
         }
         catch (Throwable t)
         {
            log.error(" a problem occurred with remote rollback", t);
            remoteRollbackException = t;
         }

      }
      Object retval = nextInterceptor(ctx);
      if (remoteRollbackException != null)
      {
         throw remoteRollbackException;
      }
      return retval;
   }

   @Override
   protected Object handlePutForExternalReadVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, DataVersion dv) throws Throwable
   {
      return handlePutForExternalReadMethod(ctx, gtx, fqn, key, value);
   }

   @Override
   protected Object handlePutForExternalReadMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value) throws Throwable
   {
      gtx = getGlobalTransaction(ctx);
      cache.getTransactionTable().get(gtx).setForceAsyncReplication(true);
      return nextInterceptor(ctx);
   }


   private GlobalTransaction getGlobalTransaction(InvocationContext ctx)
   {
      // get the current gtx
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      if (gtx == null)
      {
         throw new CacheException("failed to get global transaction");
      }
      return gtx;
   }

   protected void broadcastPrepare(MethodCall methodCall, GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.getCacheMode() == Configuration.CacheMode.REPL_SYNC;

      Object[] args = methodCall.getArgs();
      List modifications = (List) args[1];
      int num_mods = modifications != null ? modifications.size() : 0;

      // this method will return immediately if we're the only member
      if (cache.getMembers() != null && cache.getMembers().size() > 1)
      {
         // Map method calls to data versioned equivalents.
         // See JBCACHE-843 and docs/design/DataVersioning.txt
         MethodCall toBroadcast = mapDataVersionedMethodCalls(methodCall, getTransactionWorkspace(gtx));

         //record the things we have possibly sent
         broadcastTxs.add(gtx);
         if (log.isDebugEnabled())
         {
            log.debug("(" + cache.getLocalAddress()
                  + "): broadcasting prepare for " + gtx
                  + " (" + num_mods + " modifications");
         }

         replicateCall(ctx, toBroadcast, remoteCallSync, ctx.getOptionOverrides());
      }
      else
      {
         //no members, ignoring
         if (log.isDebugEnabled())
         {
            log.debug("(" + cache.getLocalAddress()
                  + "):not broadcasting prepare as members are " + cache.getMembers());
         }
      }
   }


   protected void broadcastCommit(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.isSyncCommitPhase();

      // Broadcast commit() to all members (exclude myself though)
      if (cache.getMembers() != null && cache.getMembers().size() > 1)
      {
         try
         {
            broadcastTxs.remove(gtx);
            MethodCall commit_method = MethodCallFactory.create(MethodDeclarations.commitMethod_id, gtx);

            if (log.isDebugEnabled())
               log.debug("running remote commit for " + gtx + " and coord=" + cache.getLocalAddress());

            // for an optimistic commit we don't need to force an OOB message since O/L means we have non-blocking reads.
            replicateCall(ctx, commit_method, remoteCallSync, ctx.getOptionOverrides(), false);
         }
         catch (Exception e)
         {
            log.error("Commit failed", e);
            throw e;
         }
      }
   }

   protected void broadcastRollback(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.isSyncRollbackPhase();

      if (cache.getMembers() != null && cache.getMembers().size() > 1)
      {
         // Broadcast rollback() to all other members (excluding myself)
         try
         {
            broadcastTxs.remove(gtx);
            MethodCall rollback_method = MethodCallFactory.create(MethodDeclarations.rollbackMethod_id, gtx);

            if (log.isDebugEnabled())
               log.debug("running remote rollback for " + gtx + " and coord=" + cache.getLocalAddress());
            replicateCall(ctx, rollback_method, remoteCallSync, ctx.getOptionOverrides());
         }
         catch (Exception e)
         {
            log.error("Rollback failed", e);
            throw e;
         }
      }
   }

   private MethodCall mapDataVersionedMethodCalls(MethodCall m, TransactionWorkspace w)
   {
      Object[] origArgs = m.getArgs();
      return MethodCallFactory.create(m.getMethodId(), origArgs[0], translate((List<MethodCall>) origArgs[1], w), origArgs[2], origArgs[3], origArgs[4]);
   }

   /**
    * Translates a list of MethodCalls from non-versioned calls to versioned calls.
    */
   private List<MethodCall> translate(List<MethodCall> l, TransactionWorkspace w)
   {
      List<MethodCall> newList = new ArrayList<MethodCall>();
      for (MethodCall origCall : l)
      {
         if (MethodDeclarations.isDataGravitationMethod(origCall.getMethodId()))
         {
            // no need to translate data gravitation calls.
            newList.add(origCall);
         }
         else
         {
            Object[] origArgs = origCall.getArgs();
            // get the data version associated with this orig call.

            // since these are all crud methods the Fqn is at arg subscript 1.
            Fqn fqn = (Fqn) origArgs[origCall.getMethodId() == MethodDeclarations.moveMethodLocal_id ? 0 : 1];
            // now get a hold of the data version for this specific modification
            DataVersion versionToBroadcast = getVersionToBroadcast(w, fqn);

            // build up the new arguments list for the new call.  Identical to the original lis except that it has the
            // data version tacked on to the end.
            Object[] newArgs = new Object[origArgs.length + 1];
            System.arraycopy(origArgs, 0, newArgs, 0, origArgs.length);
            newArgs[origArgs.length] = versionToBroadcast;

            // now create a new method call which contains this data version
            MethodCall newCall = MethodCallFactory.create(MethodDeclarations.getVersionedMethodId(origCall.getMethodId()), newArgs);

            // and add it to the new list.
            newList.add(newCall);
         }
      }
      return newList;
   }

   /**
    * Digs out the DataVersion for a given Fqn.  If the versioning is explicit, it is passed as-is.  If implicit, it is
    * cloned and then incremented, and the clone is returned.
    */
   private DataVersion getVersionToBroadcast(TransactionWorkspace w, Fqn f)
   {
      WorkspaceNode n = w.getNode(f);
      if (n == null)
      {
         if (trace) log.trace("Fqn " + f + " not found in workspace; not using a data version.");
         return null;
      }
      if (n.isVersioningImplicit())
      {
         DefaultDataVersion v = (DefaultDataVersion) n.getVersion();
         if (trace)
            log.trace("Fqn " + f + " has implicit versioning.  Broadcasting an incremented version.");

         // potential bug here - need to check if we *need* to increment at all, because of Configuration.isLockParentForChildInsertRemove() 
         return v.increment();
      }
      else
      {
         if (trace) log.trace("Fqn " + f + " has explicit versioning.  Broadcasting the version as-is.");
         return n.getVersion();
      }
   }

   protected TransactionWorkspace getTransactionWorkspace(GlobalTransaction gtx) throws CacheException
   {
      OptimisticTransactionEntry transactionEntry = (OptimisticTransactionEntry) cache.getTransactionTable().get(gtx);

      if (transactionEntry == null)
      {
         throw new CacheException("unable to map global transaction " + gtx + " to transaction entry");
      }

      // try and get the workspace from the transaction
      return transactionEntry.getTransactionWorkSpace();
   }

}