package org.jboss.cache.marshall.io;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.factories.annotations.Start;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * A stream pool
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
public class ObjectStreamPool
{
   BlockingQueue<ReusableObjectOutputStream> outputStreams;
   BlockingQueue<ReusableObjectInputStream> inputStreams;
   int numOutputStreams = 1, numInputStreams = 1;
   Configuration configuration;
   Log log = LogFactory.getLog(ObjectStreamPool.class);

   public ObjectStreamPool()
   {
      // for construction by the IOC framework
   }

   @Inject
   void injectDependencies(Configuration configuration)
   {
      this.configuration = configuration;
   }

   @Start
   void start() throws InterruptedException, IOException
   {
      if (configuration != null)
      {
         numInputStreams = configuration.getObjectInputStreamPoolSize();
         numOutputStreams = configuration.getObjectOutputStreamPoolSize();
         if (log.isDebugEnabled())
            log.debug("Using stream sizes from configuration as input: " + numInputStreams + " and output: " + numOutputStreams);
      }

      outputStreams = new LinkedBlockingQueue<ReusableObjectOutputStream>();
      inputStreams = new LinkedBlockingQueue<ReusableObjectInputStream>();

      long start = 0;
      if (log.isDebugEnabled())
      {
         log.debug("Creating " + numOutputStreams + " ReusableObjectOutputStreams and " + numInputStreams + " ReusableObjectInputStreams for pool");
         start = System.currentTimeMillis();
      }
      for (int i = 0; i < numOutputStreams; i++) outputStreams.put(new ReusableObjectOutputStream());
      for (int i = 0; i < numInputStreams; i++) inputStreams.put(new ReusableObjectInputStream());
      if (log.isDebugEnabled())
      {
         log.debug("Initialised object stream pool in " + (System.currentTimeMillis() - start) + " millis");
      }

   }

   /**
    * @return the next available ObjectOutputStream
    * @throws InterruptedException
    */
   public ReusableObjectOutputStream getOutputStream() throws InterruptedException
   {
      ReusableObjectOutputStream roos = outputStreams.take();
      roos.init();
      return roos;
   }

   public ObjectInputStream getInputStream(byte[] buf) throws IOException, InterruptedException
   {
//      return new MarshalledValueInputStream(new ByteArrayInputStream(buf));
      ReusableObjectInputStream rois = inputStreams.take();
      rois.init(buf);
      return rois;
   }

   public void returnStreamToPool(ReusableObjectOutputStream stream) throws IOException, InterruptedException
   {
      stream.reset();
      outputStreams.put(stream);
   }

   public void returnStreamToPool(ObjectInputStream stream) throws InterruptedException, IOException
   {
      stream.reset();
      inputStreams.put((ReusableObjectInputStream) stream);
   }
}
