package org.jboss.cache.marshall.io;

import org.jboss.cache.CacheException;
import org.jboss.util.stream.MarshalledValueInputStream;

import java.io.ByteArrayInputStream;
import java.io.IOException;

/**
 * A reusable ObjectInputStream that uses an internal byte array.  The byte array can be set using {@link #init(byte[])},
 * objects read from the byte array, and the stream reset using {@link #reset()}.
 * <p/>
 * These features allow the stream to be reused and pooled.
 * <p/>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
public class ReusableObjectInputStream extends MarshalledValueInputStream
{
   public final static byte[] INIT_BYTES = {(byte) ((STREAM_MAGIC >>> 8) & 0xFF),
         (byte) ((STREAM_MAGIC) & 0xFF),
         (byte) ((STREAM_VERSION >>> 8)),
         (byte) ((STREAM_VERSION) & 0xFF),
         TC_BLOCKDATA, 0};
   ResettableByteArrayInputStream bytes;

   public ReusableObjectInputStream() throws IOException, SecurityException
   {
      this(new ResettableByteArrayInputStream());
   }

   protected ReusableObjectInputStream(ResettableByteArrayInputStream in) throws IOException
   {
      super(in);
      bytes = in;
   }

   /**
    * Resets the ObjectInputStream so it can be reused, by resetting and clearing any internal caches and by clearing the
    * underlying byte stream to free up memory.
    *
    * @throws IOException
    */
   @Override
   public void reset() throws IOException
   {
      bytes.reset();
      bytes.clear();
   }

   /**
    * Initialises the stream to read from the byte buffer passed in
    *
    * @param buf byte buffer to read from
    */
   public void init(byte[] buf) throws IOException
   {
      bytes.init(buf);
   }

   /**
    * A byte array input stream that can be reused (i.e., have it's byte array re-initialised).
    */
   static class ResettableByteArrayInputStream extends ByteArrayInputStream
   {
      /**
       * Creates a new instance with a null byte buffer.  Use {@link #init(byte[])} to set the byte buffer to use.
       */
      ResettableByteArrayInputStream()
      {
         super(INIT_BYTES);
      }

      /**
       * Sets the underlying byte buffer to null, to free up memory after use.
       */
      public void clear()
      {
         buf = null;
      }

      /**
       * Initialises the byte stream with the byte array passed in.
       *
       * @param b byte array to use
       */
      public void init(byte[] b)
      {
         if (b.length < 4) throw new CacheException("No header bytes!");
         buf = b;
         count = buf.length;
         pos = 4;
      }
   }

}
