/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.transaction;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.marshall.MethodCall;

import javax.transaction.Transaction;
import java.util.Collection;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Maintains the mapping between a local {@link Transaction} and a {@link GlobalTransaction}.
 * Also stores {@link TransactionEntry} instances under a given transaction.
 *
 * @author <a href="mailto:bela@jboss.org">Bela Ban</a> Apr 14, 2003
 * @version $Revision: 5190 $
 */
public class TransactionTable
{

   /**
    * Mapping between local (javax.transaction.Transaction)
    * and GlobalTransactions.
    * New: a local TX can have a number of GTXs
    */
   protected Map<Transaction, GlobalTransaction> tx_map = new ConcurrentHashMap<Transaction, GlobalTransaction>();

   /**
    * Mappings between GlobalTransactions and modifications.
    */
   protected Map<GlobalTransaction, TransactionEntry> txs = new ConcurrentHashMap<GlobalTransaction, TransactionEntry>();

   /**
    * our logger
    */
   private static final Log log = LogFactory.getLog(TransactionTable.class);

   /**
    * Returns the number of local transactions.
    */
   public int getNumLocalTransactions()
   {
      return tx_map.size();
   }

   /**
    * Returns the number of global transactions.
    */
   public int getNumGlobalTransactions()
   {
      return txs.size();
   }

   /**
    * Returns the global transaction associated with the local transaction.
    * Returns null if tx is null or it was not found.
    */
   public GlobalTransaction get(Transaction tx)
   {
      if (tx == null)
         return null;
      return tx_map.get(tx);
   }

   /**
    * Returns the local transaction associated with a GlobalTransaction. Not
    * very efficient as the values have to be iterated over, don't use
    * frequently
    *
    * @param gtx The GlobalTransaction
    * @return Transaction. The local transaction associated with a given
    *         GlobalTransaction). This will be null if no local transaction is
    *         associated with a given GTX
    */
   public Transaction getLocalTransaction(GlobalTransaction gtx)
   {
      Transaction local_tx;
      GlobalTransaction global_tx;

      if (gtx == null)
         return null;
      for (Map.Entry<Transaction, GlobalTransaction> entry : tx_map.entrySet())
      {
         local_tx = entry.getKey();
         global_tx = entry.getValue();
         if (gtx.equals(global_tx))
         {
            return local_tx;
         }
      }
      return null;
   }

   /**
    * Associates the global transaction with the local transaction.
    */
   public void put(Transaction tx, GlobalTransaction gtx)
   {
      if (tx == null)
      {
         log.error("key (Transaction) is null");
         return;
      }
      tx_map.put(tx, gtx);
   }

   /**
    * Returns the local transaction entry for the global transaction.
    * Returns null if tx is null or it was not found.
    */
   public TransactionEntry get(GlobalTransaction gtx)
   {
      return gtx != null ? txs.get(gtx) : null;
   }

   /**
    * Associates the global transaction with a transaction entry.
    */
   public void put(GlobalTransaction tx, TransactionEntry entry)
   {
      if (tx == null)
      {
         log.error("key (GlobalTransaction) is null");
         return;
      }
      txs.put(tx, entry);
   }

   /**
    * Removes a global transation, returns the old transaction entry.
    */
   public TransactionEntry remove(GlobalTransaction tx)
   {
      if (tx == null) return null;
      return txs.remove(tx);
   }

   /**
    * Removes a local transation, returns the global transaction entry.
    */
   public GlobalTransaction remove(Transaction tx)
   {
      if (tx == null)
         return null;
      return tx_map.remove(tx);
   }

   /**
    * Adds a motification to the global transaction.
    */
   public void addModification(GlobalTransaction gtx, MethodCall m)
   {
      TransactionEntry entry = get(gtx);
      if (entry == null)
      {
         log.error("transaction not found (gtx=" + gtx + ")");
         return;
      }
      entry.addModification(m);
   }

   public void addCacheLoaderModification(GlobalTransaction gtx, MethodCall m)
   {
      if (m != null)
      {
         TransactionEntry entry = get(gtx);
         if (entry == null)
         {
            log.error("transaction not found (gtx=" + gtx + ")");
            return;
         }
         entry.addCacheLoaderModification(m);
      }
   }

   /**
    * Adds an undo operation to the global transaction.
    */
   public void addUndoOperation(GlobalTransaction gtx, MethodCall m)
   {
      TransactionEntry entry = get(gtx);
      if (entry == null)
      {
         log.error("transaction not found (gtx=" + gtx + ")");
         return;
      }
      entry.addUndoOperation(m);
   }

   /**
    * Adds a lock to the global transaction.
    */
   public void addLock(GlobalTransaction gtx, NodeLock l)
   {
      TransactionEntry entry = get(gtx);
      if (entry == null)
      {
         throw new CacheException("Unable to record lock for transaction " + gtx + " since no transaction entry exists!");
      }
      entry.addLock(l);
   }

   /**
    * Adds a collection of locks to the global transaction.
    */
   public void addLocks(GlobalTransaction gtx, Collection<NodeLock> locks)
   {
      TransactionEntry entry = get(gtx);
      if (entry == null)
      {
         log.error("transaction entry not found for (gtx=" + gtx + ")");
         return;
      }
      entry.addLocks(locks);
   }

   public void cleanup(GlobalTransaction gtx)
   {
      if (log.isTraceEnabled()) log.trace("Cleaning up locks for gtx " + gtx);
      TransactionEntry entry = this.get(gtx);
      // Let's do it in stack style, LIFO
      if (entry != null)
         entry.releaseAllLocksLIFO(gtx);
      else
         log.error("No transaction entry present!!", new Throwable());
   }


   /**
    * Returns summary debug information.
    */
   public String toString()
   {
      StringBuffer sb = new StringBuffer();
      sb.append(tx_map.size()).append(" mappings, ");
      sb.append(txs.size()).append(" transactions");
      return sb.toString();
   }

   /**
    * Returns detailed debug information.
    */
   public String toString(boolean print_details)
   {
      if (!print_details)
         return toString();
      StringBuffer sb = new StringBuffer();
      sb.append("LocalTransactions: ").append(tx_map.size()).append("\n");
      sb.append("GlobalTransactions: ").append(txs.size()).append("\n");
      sb.append("tx_map:\n");
      for (Map.Entry<Transaction, GlobalTransaction> entry : tx_map.entrySet())
      {
         sb.append(entry.getKey()).append(": ").append(entry.getValue()).append("\n");
      }
      sb.append("txs:\n");
      for (Map.Entry<GlobalTransaction, TransactionEntry> entry : txs.entrySet())
      {
         sb.append(entry.getKey()).append(": ").append(entry.getValue()).append("\n");
      }
      return sb.toString();
   }

   /**
    * Add the lock to the list of locks maintained for this transaction
    * (needed for release of locks on commit or rollback)
    */
   public void recordNodeLock(GlobalTransaction gtx, NodeLock lock)
   {
      try
      {
         addLock(gtx, lock);
      }
      catch (CacheException e)
      {
         // may happen, if the transaction entry does not exist
         lock.release(gtx);
         throw e;
      }
   }

   public boolean isNodeRemovedInTx(GlobalTransaction gtx, Fqn fqn)
   {
      TransactionEntry te = get(gtx);
      return te != null && te.getRemovedNodes().contains(fqn);
   }
}
