package org.jboss.cache.marshall.io;

import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.Date;
import java.util.Random;

/**
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
@Test(groups = "functional")
public class ObjectStreamPoolTest implements Serializable
{
   transient ObjectStreamPool pool;
   transient ReusableObjectOutputStream oos;
   transient ObjectInputStream ois;

   @BeforeTest
   public void setUp() throws IOException, InterruptedException
   {
      pool = new ObjectStreamPool();
      pool.numInputStreams = 1;
      pool.numOutputStreams = 1;
      pool.start();
   }

   @AfterMethod
   public void afterMethod() throws IOException, InterruptedException
   {
      // ensure we return stuff to the pool
      if (oos != null) pool.returnStreamToPool(oos);
      if (ois != null) pool.returnStreamToPool(ois);
   }

   public void testStreamReset() throws IOException, ClassNotFoundException, InterruptedException
   {
      oos = pool.getOutputStream();

      Object toWrite = "BlahBlah";
      oos.writeObject(toWrite);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      toWrite = new Integer(9);
      oos.writeObject(toWrite);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      toWrite = new Date();
      oos.writeObject(toWrite);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      toWrite = new MyClass(876, "Hello");
      oos.writeObject(toWrite);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
   }

   /**
    * To test block data header sizes
    *
    * @throws IOException
    * @throws ClassNotFoundException
    * @throws InterruptedException
    */
   public void testLargeStreams() throws IOException, ClassNotFoundException, InterruptedException
   {
      oos = pool.getOutputStream();

      StringBuilder sb = new StringBuilder();
      Random r = new Random();
      // 1K string
      for (int i = 0; i < 0xFFFF; i++) sb.append((char) (r.nextInt(26) + 65));

      Object toWrite = sb.toString();
      System.out.println("Writing string (length = " + ((String) toWrite).length() + ") " + toWrite);
      oos.writeObject(toWrite);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      toWrite = new Integer(9);
      for (int i = 0; i < 0xFFFF; i++) oos.writeObject(toWrite);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      for (int i = 0; i < 0xFFFF; i++) assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      toWrite = new MyClass(1, "x");
      for (int i = 0; i < 0xFFFF; i++) oos.writeObject(toWrite);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      for (int i = 0; i < 0xFFFF; i++) assert ois.readObject().equals(toWrite);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();
      int arraySize = 0xFFFF;
      Object[] toWriteArray = new Object[arraySize];
      for (int i = 0; i < arraySize; i++)
      {
         toWriteArray[i] = new MyClass(i, "x");
         oos.writeObject(toWriteArray[i]);
      }
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      for (int i = 0; i < arraySize; i++)
      {
         Object read = ois.readObject();
         assert read.equals(toWriteArray[i]) : "Loop " + i + ": Expected " + toWriteArray[i] + " but was " + read;
      }
   }

   public void testStreamResetWithFlush() throws Exception
   {
      oos = pool.getOutputStream();

      Object toWrite = "BlahBlah", toWrite2 = "Hello world";
      oos.writeObject(toWrite);
      oos.flush();
      oos.writeObject(toWrite2);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
      assert ois.readObject().equals(toWrite2);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();

      toWrite = new Integer(9);
      toWrite2 = new Long(99);
      oos.writeObject(toWrite);
      oos.flush();
      oos.writeObject(toWrite2);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
      assert ois.readObject().equals(toWrite2);

      // return to pool
      pool.returnStreamToPool(oos);
      pool.returnStreamToPool(ois);

      oos = pool.getOutputStream();

      toWrite = new Date();
      toWrite2 = Boolean.FALSE;
      oos.writeObject(toWrite);
      oos.flush();
      oos.writeObject(toWrite2);
      asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
      assert ois.readObject().equals(toWrite2);
   }

   public void testCustomTypes() throws Exception
   {
      oos = pool.getOutputStream();

      MyClass toWrite = new MyClass(123, "hello");
      MyClass toWrite2 = new MyClass(456, "world");
      oos.writeObject(toWrite);
      oos.flush();
      oos.writeObject(toWrite2);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
      assert ois.readObject().equals(toWrite2);
   }

   public void testMixedTypes() throws Exception
   {
      oos = pool.getOutputStream();

      MyClass toWrite = new MyClass(123, "hello");
      MyClass toWrite2 = new MyClass(456, "world");
      oos.writeObject(toWrite);
      oos.flush();
      oos.writeBoolean(false);
      oos.flush();
      oos.writeObject(toWrite2);
      oos.flush();
      oos.writeShort(6);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);
      assert ois.readObject().equals(toWrite);
      assert ois.readBoolean() == false;
      assert ois.readObject().equals(toWrite2);
      assert ois.readShort() == 6;
   }

   public void testLargeObject() throws Exception
   {
      oos = pool.getOutputStream();
      StringBuilder builder = new StringBuilder();
      int targetSize = 1 << 22; // 4MB  !!
      System.out.println("target size: " + targetSize);
      for (int i = 0; i < targetSize; i++) builder.append("X");
      String fourMegString = builder.toString();
      oos.writeObject(fourMegString);
      byte[] asBytes = oos.getBytes();

      ois = pool.getInputStream(asBytes);

      assert ois.readObject().equals(fourMegString);
   }

   public void testReusableOutputStreamArraySize() throws InterruptedException, IOException
   {
      oos = pool.getOutputStream();
      int originalLength = oos.baos.getBufSize();
      for (int i = 0; i < 0xFF; i++) oos.write(7);
      oos.flush();
      assert oos.baos.getBufSize() > 0xFF;

      oos.reset();
      oos.init();
      assert oos.baos.getBufSize() == originalLength;
   }

   public static class MyClass implements Serializable
   {
      int number;
      String name;

      public MyClass(int number, String name)
      {
         this.number = number;
         this.name = name;
      }

      public boolean equals(Object o)
      {
         if (this == o) return true;
         if (o == null || getClass() != o.getClass()) return false;

         MyClass myClass = (MyClass) o;

         if (number != myClass.number) return false;
         if (name != null ? !name.equals(myClass.name) : myClass.name != null) return false;

         return true;
      }

      public int hashCode()
      {
         int result;
         result = number;
         result = 31 * result + (name != null ? name.hashCode() : 0);
         return result;
      }

      public String toString()
      {
         return "MyClass{" +
               "number=" + number +
               ", name='" + name + '\'' +
               '}';
      }
   }
}
